`timescale 1 ns/ 1 ps
module bit_align_ctl
(input        rx_clk
,input        reset
,input        bit_aln_en
,input  [6:0] rxclk_word
,output reg   phase_dir
,output reg   phase_rot
,output reg   dphase_lock
,output reg [3:0] dphase        
);
        
parameter IDLE    = 2'b00, 
          D_PHASE = 2'b01,
          JUDGE   = 2'b10,
          LOCK    = 2'b11;

parameter INIT_PHASE = 1'b0; // 1'b1 for 180C, 1'b0 for 0C. It is used to match delay between ECLK and SCLK.
                             // In general, 180C for ECLK connecting dedicate edge clock net only. e.g.,  
                             // LVDS input signals from FPGA left or right side. If ECLK also uses other generic
                             // routing net, 180C or 0C can be tried to get the best PLL phase.
        
parameter DP_CNT_WIDTH   = 20; 
parameter WORK_BIT       = 12; 
parameter DATA_CNT_WIDTH = 8;     
parameter STABLE_THD = 8'h80;        
parameter CHANGE_THD = 8'h00;       
   
reg [5:0] rot_cnt;            
reg [1:0] dphase_state;
   
reg [6:0]                rxclk_word_buf;
reg [6:0]                cur_stable_rxclk_word ;
reg                      data_change_cnt_en;
reg [DP_CNT_WIDTH-1:0]   dphase_cnt;
reg [2:0]                dphase_cnt_h_buf;
reg                      dphase_cnt_msb_buf;
reg                      data_stable_start;
reg                      data_unstable;
reg [7:0]                bit_valid; // correspoding to 8 phases
reg [DATA_CNT_WIDTH-1:0] data_change_cnt;   
reg [DATA_CNT_WIDTH-1:0] data_stable_cnt;
reg [3:0]                judge_cnt;
reg [7:0]                sf_bit_valid; // correspoding to 8 phases
reg [3:0]                stable_bit_num;   
reg [2:0]                unst_to_st_num;   
reg [2:0]                unst_to_st_start_bit; 
   
wire data_change;
wire cur_stable_rxclk_word_err;
 
always @(posedge rx_clk or posedge reset)
begin
   if (reset) begin
      dphase_cnt <= 0;
      data_change_cnt_en <= 1'b0;
      dphase_cnt_h_buf <= 3'b0;
      dphase_cnt_msb_buf <= 1'b0;//
   end else if (bit_aln_en) begin
      if (dphase_cnt[DP_CNT_WIDTH-1] == 1'b0) 
         dphase_cnt <= dphase_cnt + 1;
      else if (dphase_state == IDLE)
         dphase_cnt <= 0;        
      
      if (dphase_cnt[WORK_BIT] == 1'b1)
         data_change_cnt_en <= 1'b1;
      else if (dphase_cnt[DP_CNT_WIDTH-5 : 0] == 0 && 
               dphase_state == D_PHASE) //Keep data_change_cnt_en = 1'b1 after D_PHASE state
         data_change_cnt_en <= 1'b0;   
         
      dphase_cnt_h_buf <= dphase_cnt[DP_CNT_WIDTH-2 : DP_CNT_WIDTH-4];   
      dphase_cnt_msb_buf <= dphase_cnt[DP_CNT_WIDTH-1];
   end
end
   
assign data_change = rxclk_word !== rxclk_word_buf ? 1'b1 : 1'b0;

assign cur_stable_rxclk_word_err = (data_stable_start == 1'b0 ||
                                    cur_stable_rxclk_word == 7'b1100011 ||
                                    cur_stable_rxclk_word == 7'b1000111 ||
                                    cur_stable_rxclk_word == 7'b0001111 ||
                                    cur_stable_rxclk_word == 7'b0011110 ||
                                    cur_stable_rxclk_word == 7'b0111100 ||
                                    cur_stable_rxclk_word == 7'b1111000 ||
                                    cur_stable_rxclk_word == 7'b1110001) ?
                                    1'b0 : 1'b1;

always @(posedge rx_clk or posedge reset)
begin
   if (reset) begin
      phase_dir <= 1'b0;
      phase_rot <= 1'b1;
      rot_cnt <= 6'd0;
      dphase_lock <= 1'b0;
      rxclk_word_buf <= 0;
      data_stable_cnt <= 0;
      data_stable_start <= 1'b0;
      data_unstable <= 1'b0;
      cur_stable_rxclk_word <= 0;
      data_change_cnt <= 0;
      dphase <= INIT_PHASE ? 4'b1000 : 4'b0000;
      bit_valid <= 0;
   end else if (bit_aln_en == 1'b1) begin
      rxclk_word_buf <= rxclk_word;
      
      if (data_change_cnt_en == 1'b0 || data_change == 1'b1) 
         data_stable_cnt <= 0;
      else if (data_stable_cnt[DATA_CNT_WIDTH-1] == 1'b0)
         data_stable_cnt <= data_stable_cnt + 1;
      
      if (data_change_cnt_en == 1'b0) begin
         data_stable_start <= 1'b0;
         data_unstable <= 1'b0;
      end else if (data_stable_cnt >= STABLE_THD && data_change !==1'b1) begin  
         data_stable_start <= 1'b1;
         
         if (cur_stable_rxclk_word !== rxclk_word && data_stable_start == 1'b1) 
            data_unstable <= 1'b1;
             
         cur_stable_rxclk_word <= rxclk_word;
         
      end
      
      if (data_change_cnt_en == 1'b0) 
         data_change_cnt <= 0;
      else if (data_change == 1'b1 && data_change_cnt[DATA_CNT_WIDTH-1] == 1'b0)
         data_change_cnt <= data_change_cnt + 1;

      if (dphase_cnt[DP_CNT_WIDTH-1] == 1'b0) begin
         dphase[3] <= INIT_PHASE;
         dphase[2:0] <= dphase_cnt[DP_CNT_WIDTH-2 : DP_CNT_WIDTH-4];         
         
         rot_cnt <= 6'd0;
         phase_dir <= 1'b0;
         
         if ((dphase_cnt_h_buf[0] ^ dphase_cnt[DP_CNT_WIDTH-4]) || dphase_cnt_msb_buf)
            phase_rot <= 1'b0;
         else if (dphase_cnt[2])
            phase_rot <= 1'b1;
            
         dphase_lock <= 1'b0;   
      end else if (dphase_state == LOCK) begin
         dphase[3] <= INIT_PHASE ? ~ dphase_cnt[DP_CNT_WIDTH-2] : dphase_cnt[DP_CNT_WIDTH-2];
         dphase[2:0] <= unst_to_st_start_bit + stable_bit_num[3:1];
         
         phase_dir <= 1'b1;
         
         if (dphase[2:0] != rot_cnt[5:3]) begin
            rot_cnt <= rot_cnt - 1;
            
            if (rot_cnt[2:0] == 3'b110)
               phase_rot <= 1'b0;
            else if (rot_cnt[2:0] == 3'b010)   
               phase_rot <= 1'b1;
               
         end else 
            dphase_lock <= 1'b1;
      end
      
      if (dphase_state == IDLE)
         bit_valid <= {8{1'b1}};
      else if (data_change_cnt_en == 1'b1 && 
               (data_change_cnt > CHANGE_THD || data_unstable == 1'b1 || cur_stable_rxclk_word_err == 1'b1))
         bit_valid[dphase_cnt_h_buf] <= 1'b0;
                                 
   end
end


always @(posedge rx_clk or posedge reset)
begin
   if (reset == 1'b1) begin
      dphase_state <= IDLE;
      judge_cnt <= 4'b0000;
      sf_bit_valid <= 0;
      stable_bit_num <= 4'b0000;
      unst_to_st_num <= 3'b000;
      unst_to_st_start_bit <= 3'b000;
   end else if (bit_aln_en == 1'b1)
      case (dphase_state) 
         IDLE    : begin
                      if (dphase_cnt[0] == 1'b1) 
                         dphase_state <= D_PHASE;
                      
                      judge_cnt <= 4'b0000;                      
                   end
                      
         D_PHASE : begin
                      if (dphase_cnt[DP_CNT_WIDTH-1] == 1'b1)
                         dphase_state <= JUDGE;
                      
                      sf_bit_valid <= bit_valid;
                      stable_bit_num <= 4'b0000;
                      unst_to_st_num <= 3'b000;
                      unst_to_st_start_bit <= 3'b000;
                   end
                   
         JUDGE   : begin
                      judge_cnt <= judge_cnt + 1;
                 
                      if (judge_cnt[3] == 1'b0) begin
                         sf_bit_valid <= {sf_bit_valid[0], sf_bit_valid[7:1]};
                         
                         if (sf_bit_valid[0] == 1'b1)
                            stable_bit_num <= stable_bit_num + 1;
                         
                         if (sf_bit_valid[1:0] == 2'b10) begin
                            unst_to_st_num <= unst_to_st_num + 1;
                            unst_to_st_start_bit <= judge_cnt[2:0] + 1;
                         end
                      end
                                      
                      if (judge_cnt[3] == 1'b1) begin
                         if (stable_bit_num == 4'b0000 || (unst_to_st_num !== 3'b001 && stable_bit_num[3] == 1'b0)) 
                            dphase_state <= IDLE;
                         else begin
                            dphase_state <= LOCK;    
                            stable_bit_num <= stable_bit_num - 1;
                         end   
                      end   
                   end
                   
         LOCK    : begin
                      dphase_state <= LOCK;
                   end      
      endcase                                                         
end

endmodule